package org.processmining.xlu.matching.plugins.alg.impl;

import gnu.trove.set.hash.THashSet;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.impl.MatchMappingImpl;
import org.processmining.xlu.matching.plugins.alg.CostStrategy;
import org.processmining.xlu.matching.plugins.alg.Criterion;
import org.processmining.xlu.matching.plugins.alg.HeuristicStrategy;
import org.processmining.xlu.matching.plugins.alg.MatchingAlg;

public abstract class MatchingAlgAbstract implements MatchingAlg {

	private CostStrategy costfunction;
	private List<Criterion> criteria;
	private HeuristicStrategy heuristicfunction;

	private Set<String> labels;

	public HeuristicStrategy getHeuristicfunction() {
		return heuristicfunction;
	}

	public void setHeuristicfunction(HeuristicStrategy heuristicfunction) {
		this.heuristicfunction = heuristicfunction;
	}

	public String toString() {
		return this.getClass().getSimpleName();
	}

	protected MatchingAlgAbstract() {
		criteria = new ArrayList<Criterion>();
		labels = new THashSet<String>();
	}

	public void setMatchingLabelSet(Set<String> labels) {
		this.labels = labels;
	}

	public void setCostFunction(CostStrategy costfunc) {
		this.costfunction = costfunc;
	}

	public void setCriteria(List<Criterion> criteria) {
		this.criteria = criteria;
	}

	public CostStrategy getCostStrategy() {
		return this.costfunction;
	}

	public List<Criterion> getCriteria() {
		return this.criteria;
	}

	public abstract MatchMapping getOptimalMapping(EGraph g1, EGraph g2);

	protected boolean isValid(MatchMapping mapping) {
		boolean r = true;
		for (Criterion c : getCriteria()) { // if no criteria, then true
			r &= c.isValid(mapping);
		}
		return r;
	}

	protected List<Integer> getNodesAllowedToMap(ENode source, List<ENode> targets) {
		List<Integer> ids = new ArrayList<Integer>();
		for (int i = 0; i < targets.size(); i++) {
			if (targets.get(i).getLabel().equals(source.getLabel())) { // get all nodes that have the same label as source. 
				ids.add(i);
			}
		}
		return ids;
	}

	protected void matchDummyStartEndAndUpdateMapping(Collection<ENode> sources,
			Collection<ENode> targets, MatchMapping mapping) {
		//Match start and end remove from choices. 
		List<ENode> invSources = getInvisibleNodesAndRemoveFrom(sources);
		List<ENode> invTarget = getInvisibleNodesAndRemoveFrom(targets);
		for (ENode e : invSources) {
			for (ENode t : invTarget) {
				if (e.getLabel().equals(t.getLabel())) {
					mapping.addMatch(e, t);
					break;
				}
			}
		}
		matchOtherLabels(sources, targets, mapping);
	}

	protected void matchOtherLabels(Collection<ENode> sources, Collection<ENode> targets,
			MatchMapping mapping) {
		if (this.labels == null || this.labels.isEmpty()) {
			return;
		}

		List<ENode> autoMatchSources = getMatchNodesAndRemoveFrom(sources, this.labels);
		List<ENode> autoMatchTarget = getMatchNodesAndRemoveFrom(targets, this.labels);
		for (ENode e : autoMatchSources) {
			ENode match = null;
			for (ENode t : autoMatchTarget) {
				if (e.getLabel().equals(t.getLabel())) {
					match = t;
					break;
				}
			}
			if (match != null) {
				autoMatchTarget.remove(match);
			}
			mapping.addMatch(e, match);
		}
		for (ENode t : autoMatchTarget) {
			mapping.addMatch(null, t);
		}
	}

	private List<ENode> getMatchNodesAndRemoveFrom(Collection<ENode> sources,
			Set<String> shouldBeMatchLabelsByAlg) {
		List<ENode> resToBeAutoMatched = new ArrayList<ENode>();
		for (ENode e : sources) {
			if (!shouldBeMatchLabelsByAlg.contains(e.getLabel())) {
				resToBeAutoMatched.add(e);
			}
		}
		sources.removeAll(resToBeAutoMatched);
		return resToBeAutoMatched;
	}

	private List<ENode> getInvisibleNodesAndRemoveFrom(Collection<ENode> sources) {
		List<ENode> res = new ArrayList<ENode>();
		for (ENode e : sources) {
			if (!e.isVisible()) {
				res.add(e);
			}
		}
		sources.removeAll(res);
		return res;
	}

	protected void addMappingOnlyIfOpt(List<MatchMapping> mappings, MatchMapping mapping,
			int cost) {
		int optCost = Integer.MAX_VALUE;
		// update opt
		if (mappings.size() > 0) {
			optCost = mappings.get(0).getCost();
		}
		if (cost <= optCost) {
			if (cost < optCost) {
				mappings.clear();
			}
			cloneMappingAndAddTo(mappings, mapping, cost);
		}
	}

	protected void cloneMappingAndAddTo(List<MatchMapping> mappings,
			MatchMapping mapping, int cost) {
		MatchMapping newMapping = new MatchMappingImpl(mapping.getSourceGraph(),
				mapping.getTargetGraph());
		newMapping.clone(mapping);
		newMapping.setCost(cost);
		mappings.add(newMapping);
	}
}
