package org.processmining.xlu.matching.plugins.alg.impl;

import java.util.ArrayList;
import java.util.List;

import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.ENode;
import org.processmining.xlu.matching.models.MatchMapping;
import org.processmining.xlu.matching.models.impl.MatchMappingImpl;
import org.processmining.xlu.matching.plugins.alg.MatchingAllAlg;

/**
 * This matching algorithm do still have full recursion but can use a heuristic
 * function to prune search state-space
 * 
 * @author xlu
 * 
 */
public class MatchingAlgBackTrackingPruningImpl extends MatchingAlgAbstract implements MatchingAllAlg{


	public MatchMapping getOptimalMapping(EGraph g1, EGraph g2) {
		List<MatchMapping> maps =  getAllOptimalMapping(g1, g2);
		return maps.get(0);
	}
	
	public List<MatchMapping> getAllOptimalMapping(EGraph g1, EGraph g2) {
		MatchMapping mapping = new MatchMappingImpl(g1, g2);
		List<MatchMapping> optMappings = new ArrayList<MatchMapping>();

		List<ENode> sources = new ArrayList<ENode>(g1.getNodes());
		List<ENode> targets = new ArrayList<ENode>(g2.getNodes());

		matchDummyStartEndAndUpdateMapping(sources, targets, mapping);
		recursePruning(sources, targets, mapping, 0, optMappings, Integer.MAX_VALUE);
		
		

		return optMappings;
	}

	protected int recursePruning(List<ENode> sources, List<ENode> targets, 
			MatchMapping mapping, Integer costSoFar,
			List<MatchMapping> optMappings, Integer minCost) {

		// 1. base case
		if (sources.isEmpty() || targets.isEmpty()) { 
			minCost = computeActualCostAndUpdateOptMapping(sources, targets, mapping, costSoFar, optMappings, minCost);
			return minCost;
		}

		// 2.1 recurse : select the source node to be mapped
		ENode source = sources.remove(0); 
		List<Integer> indicesMatchbleInTarget = getNodesAllowedToMap(source, targets);
		
		// 2.2: select the target node to be mapped. 
		for (Integer targetIndex : indicesMatchbleInTarget) { 

			ENode target = targets.remove((int) targetIndex); // select choice II. 
			// 2.3 add the choice to continue with recursion
			mapping.addMatch(source, target); 

			if (isValid(mapping)) {
				// 3. Update heuristic cost so far && Pruning!.
				int newCost = costSoFar + getHeuristicfunction().getCost(source, target, mapping);  
				if (newCost <= minCost) {
					int cost = recursePruning(sources, targets, mapping, newCost, optMappings, minCost);
					if (cost < minCost) {
						minCost = cost;
					}
				}  else {// else do not compute mapping. 	
//					System.out.println("Pruned");
				}
			}

			mapping.removeMatch(source, target);
			targets.add(targetIndex, target); // add element back to original index.
		}

		// option map to /nothing...
		mapping.addMatch(source, null);
		if (isValid(mapping)) {
			int newCost = costSoFar + getHeuristicfunction().getCostNoMatch(source, mapping); // update cost so far. 
			if (newCost <= minCost) {
				int cost = recursePruning(sources, targets, mapping, newCost, optMappings, minCost);
				if (cost < minCost) {
					minCost = cost;
					//					optMapping.clone(mapping);
				}
			} else {// else do not compute mapping. 	
//				System.out.println("Pruned");
			} 	
		}
		mapping.removeMatch(source, null);
		sources.add(0, source); // undo choice. 

		return minCost;
	}

	protected int computeActualCostAndUpdateOptMapping(List<ENode> sources, List<ENode> targets, MatchMapping mapping, int costSoFar,
			List<MatchMapping> optMappings, int minCost) {
		// Complete choices
		for (ENode n : sources) {
			mapping.addMatch(n, null);
		}
		for (ENode n : targets) {
			mapping.addMatch(null, n);
		}
		
		// Compute cost and update optimal mapping
		int cost = this.getCostStrategy().getTotalCost(mapping.getSourceGraph(), mapping.getTargetGraph(), mapping);
		if(cost < costSoFar){
			System.err.println("MatchingAlgPruning : Heuristic costfunction is not the lower bound");
		}
		// update optimal mapping
		if (cost <= minCost) {
			addMappingOnlyIfOpt(optMappings, mapping, cost);
			minCost = cost;
		}
		
		// remove choices
		for (ENode n : sources) {
			mapping.removeMatch(n, null);
		}
		for (ENode n : targets) {
			mapping.removeMatch(null, n);
		}
		return minCost;
	}





}
