package org.processmining.xlu.matching.preprocessing;

import edu.uci.ics.jung.algorithms.shortestpath.UnweightedShortestPath;
import edu.uci.ics.jung.graph.DirectedSparseGraph;
import gnu.trove.map.hash.THashMap;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XEvent;
import org.processmining.partialorder.models.dependency.PDependency;
import org.processmining.partialorder.models.dependency.PDependencyImp;
import org.processmining.partialorder.ptrace.model.PTrace;
import org.processmining.xlu.matching.models.EEdge;
import org.processmining.xlu.matching.models.EGraph;
import org.processmining.xlu.matching.models.EGraphFactory;
import org.processmining.xlu.matching.models.ENode;

public abstract class PTraceToEGConverterAbstract implements PTraceToEGConverter {

	protected int contextDist = 1;
	//	protected boolean doComputeConcurrent = false;

	// classifier. 
	private XEventClasses classes;
	private Map<String, Integer> mapLabel2Index;

	private int dummyEndIndex;
	private int dummyStartIndex;

	
	protected PTraceToEGConverterAbstract(int contextDist, XEventClasses classes) {
		this.contextDist = contextDist;
		this.classes = classes;
		mapLabel2Index = new THashMap<String, Integer>(classes.size());
		for(XEventClass c : classes.getClasses()){
			mapLabel2Index.put(c.toString(), c.getIndex());
		}
		dummyStartIndex = classes.size();
		dummyEndIndex = classes.size() + 1;
		mapLabel2Index.put(EGraphFactory.Dummy_Start_Label, dummyStartIndex);
		mapLabel2Index.put(EGraphFactory.Dummy_End_Label, dummyEndIndex);
	}
	
	public XEventClasses getClasses() {
		return classes;
	}

	public void setClasses(XEventClasses classes) {
		this.classes = classes;
	}

	public Map<String, Integer> getMapNodeLabelToIndex(){
		return mapLabel2Index;
	}

	// Auxiliary functions
	protected int[][] computeNodeContexts(EGraph eg, PTrace ptrace, Map<Integer, List<Integer>> predMap,
			Map<Integer, List<Integer>> succMap, Map<Integer, List<Integer>> concMap) {
		// Initialize for each node
		int size = ptrace.getTrace().size();

		for (int i = 0; i < size + 2; i++) {
			predMap.put(i, new ArrayList<Integer>());
			succMap.put(i, new ArrayList<Integer>());
			concMap.put(i, new ArrayList<Integer>());
		}

		Integer dummyStartId = size;
		Integer dummyEndId = size + 1;

		DirectedSparseGraph<Integer, PDependency> graph = new DirectedSparseGraph<Integer, PDependency>();
		for (PDependency e : ptrace.getDependencies()) {
			if (e.isDirect()) {
				graph.addVertex(e.getSource());
				graph.addVertex(e.getTarget());
				graph.addEdge(e, e.getSource(), e.getTarget());
			}
		}
		graph.addVertex(dummyStartId);
		graph.addVertex(dummyStartId);
		for (Integer s : ptrace.getStartEventIndices()) {
			graph.addEdge(new PDependencyImp(dummyStartId, s), dummyStartId, s);
		}
		for (Integer e : ptrace.getEndEventIndices()) {
			graph.addEdge(new PDependencyImp(e, dummyEndId), e, dummyEndId);
		}

		UnweightedShortestPath<Integer, PDependency> dist = new UnweightedShortestPath<Integer, PDependency>(graph);

		// fill in distance once for all. Also fill in preds, succs and concs for each node. 
		int[][] distMatrix = new int[size + 2][size + 2];
		for (int i = 0; i < size + 2; i++) {
			for (int j = i + 1; j < size + 2; j++) {
				Number distIJ = dist.getDistance(i, j);
				Number distJI = dist.getDistance(j, i);
				if (distIJ == null && distJI == null) {
					distMatrix[i][j] = 0;
					distMatrix[j][i] = 0;
					// update contexts
					concMap.get(i).add(j);
					concMap.get(j).add(i);
				} else if (distIJ == null) { // i.e. i -/-> j and j -> i
					int length = distJI.intValue();
					distMatrix[i][j] = 0 - length;
					distMatrix[j][i] = length;
					// update contexts
					if (Math.abs(length) <= this.contextDist) {
						succMap.get(j).add(i);
						predMap.get(i).add(j);
					}

				} else if (distJI == null) {// i.e. j -/-> i and i -> j 
					int length = distIJ.intValue();
					distMatrix[i][j] = length;
					distMatrix[j][i] = 0 - length;

					if (Math.abs(length) <= this.contextDist) {
						succMap.get(i).add(j);
						predMap.get(j).add(i);
					}
				} // else error. 
			}
		}
		return distMatrix;
	}

	protected int[] convertSuccToArray(PTrace ptrace, Collection<Integer> eventIndices) {
//		XEventClasses classes = this.getClasses();
//		int dummyStartIndex = classes.size();
//		int dummyEndIndex = classes.size() + 1;

		int[] array = new int[classes.size() + 2];
		if (eventIndices.size() == 0) {
			array[dummyEndIndex] = 1;
		}

		for (Integer index : eventIndices) {
			if (index >= ptrace.size()) {
				if (index == ptrace.size()) {
					array[dummyStartIndex] += 1;
				} else {
					array[dummyEndIndex] += 1;
				}
			} else {
				XEvent e = ptrace.getEvent(index);
				XEventClass c = classes.getClassOf(e);
				array[this.mapLabel2Index.get(c.toString())] += 1; // The index of c in classes is the index of array. 
			}
		}

		return array;
	}

	protected int[] convertPredToArray(PTrace ptrace, Collection<Integer> eventIndices) {
//		XEventClasses classes = this.getClasses();
//		int dummyStartIndex = classes.size();
//		int dummyEndIndex = classes.size() + 1;

		int[] array = new int[this.mapLabel2Index.size()];
		if (eventIndices.size() == 0) {
			array[dummyStartIndex] = 1;
		}

		for (Integer index : eventIndices) {
			if (index >= ptrace.size()) {
				if (index == ptrace.size()) {
					array[dummyStartIndex] += 1;
				} else {
					array[dummyEndIndex] += 1;
				}
			} else {
				XEvent e = ptrace.getEvent(index);
				XEventClass c = classes.getClassOf(e);
				array[mapLabel2Index.get(c.toString())] += 1; // The index of c in classes is the index of array. 
			}
		}

		return array;
	}

	protected int[] convertToArray(PTrace ptrace, Collection<Integer> eventIndices) {
//		XEventClasses classes = this.getClasses();
//		int dummyStartIndex = classes.size();
//		int dummyEndIndex = classes.size() + 1;
		int[] array = new int[this.mapLabel2Index.size()];
		for (Integer index : eventIndices) {
			XEvent e = ptrace.getEvent(index);
			XEventClass c = classes.getClassOf(e);
			array[mapLabel2Index.get(c.toString())] += 1; // The index of c in classes is the index of array. 
		}

		return array;
	}

	protected void addArtificalStartAndEnd(EGraph eg, PTrace ptrace, EGraphFactory factory) {
		ENode start = factory.createArtificalStart(eg);
		ENode end = factory.createArtificalEnd(eg);

		for (Integer i : ptrace.getStartEventIndices()) {
			EEdge e = factory.createEdge(eg, start.getId(), i);
		}
		for (Integer i : ptrace.getEndEventIndices()) {
			EEdge e = factory.createEdge(eg, i, end.getId());
		}
	}
	
	public  XEventClass getClass(XEvent e){
		return classes.getClassOf(e);
	}

}
